var timeAtStart = performance.now();

var Sym3EngineState =
{
    Uninitialized: 0,
    AwaitingConnection: 1,
    AwaitingConnectionParameters: 2,
    AwaitingProjectCleared: 3,
    AwaitingProjectParts: 4,
    AwaitingProjectValidation: 5,
    AwaitingProjectInitialization: 6,
    AwaitingProjectLoadComplete: 7,
    Active: 8,
    ActiveAndRendering: 9
};
Sym3EngineState = Object.freeze(Sym3EngineState);

/** 
 * Creates and manages string backed by Emscripten memory.
 * 
 * `new Sym3String("こんにちは、世界");`
 * If called with a string, memory large enough to hold that string is
 * allocated on the Emscripten heap, and the string is copied into it,
 * along with a terminating nul character.
 * 
 * `new Sym3String(1024);`
 * If called with a number, memory large enough to hold that many
 * JavaScript characters is allocated on the Emscripten heap.
 * 
 * Call Free once you're done with the string.
 */
function Sym3String(argument)
{
    this.m_Ptr = null;
    this.m_Length = 0;  // In two byte wide characters.

    /// Create a Sym3String from a JavaScript string.
    this.AllocateFromString = function (str)
    {
        this.AllocateFromNumber(str.length + 1);
        this.CopyFromString(str);
    };

    /// Copy the contents of a JavaScript string to this string, with a terminating 0.
    this.CopyFromString = function (str)
    {
        if ((str.length + 1) > this.m_Length)
        {
            throw new Error(`Sym3String too small for string of length ${str.length}.`);
        }
        let view = this.GetView();
        for (let i = 0; i < str.length; ++i)
        {
            view[i] = str.charCodeAt(i);
        }
        view[str.length] = 0;
    };

    /// Create a Sym3String large enough to hold n JavaScript characters.
    this.AllocateFromNumber = function(n)
    {
        this.m_Ptr = Module._malloc(n * 2);
        this.m_Length = n;
    };

    /// Release the memory associated with this string.
    this.Free = function ()
    {
        Module._free(this.m_Ptr);
    };

    /**
     * Get a TypedArray view of the string.
     * Do not store this; if an allocation causes Emscripten to grow memory,
     * this view will become invalidated.
     */
    this.GetView = function()
    {
        return new Uint16Array(Module.HEAPU16.buffer, this.m_Ptr, this.m_Length);
    };

    /// Get the pointer to the memory where the string is stored.
    this.GetPtr = function()
    {
        return this.m_Ptr;
    };

    /// Get the length of this string in two byte characters.
    this.GetLength = function()
    {
        return this.m_Length;
    };

    if (typeof argument == "string")
    {
        this.AllocateFromString(argument);
    }
    else if (typeof argument == "number")
    {
        this.AllocateFromNumber(argument);
    }
    else
    {
        throw new Error(`Can't make a Sym3String from ${typeof argument}.`);
    }
}

function Sym3RuntimeAPI(clientCategories, serverName, webSocketPort)
{
    this.m_ServerName = serverName;
    this.m_WebSocketPort = webSocketPort;
    this.m_ExternalDriverGetDataToSend = 0;
    this.m_LastEngineState = 0;
    this.m_SendBuffer = null;
    this.m_SendBufferSize = 2048;
    this.m_ClientCategories = clientCategories;
    this.m_ConsoleLogsEnabled = false;

    this.m_OnProjectStateChanged = null;
    this.m_OnProjectInitialised  = null;
    this.m_OnConnectionChanged   = null;

    /* When we addFunction, if we don't removeFunction
     * eventually we'll run out of reserved function pointers. (-s RESERVED_FUNCTION_POINTERS=N when building)
     * So we keep track of functions we add so that doesn't happen.
     */
    this.m_TouchedFuncPtr = null;  ///< Holds the raw pointer for SetOnComponentTouchedCallback.
    this.m_ComponentSelectionPtr = null;  ///< Holds the raw pointer for SetOnComponentSelectionChanged.

    if (this.m_ServerName == null)
        this.m_ServerName = location.hostname;

    this.sym3Socket = new Sym3WebSocket(this.m_ServerName, webSocketPort);

    /// Subscribe to the OnComponentTouched callback.
    this.SetOnComponentTouchedCallback = function (callbackFunction)
    {
        if (this.m_TouchedFuncPtr !== null)
        {
            removeFunction(this.m_TouchedFuncPtr);
        }

        this.m_TouchedFuncPtr = addFunction(function (lowComponentId, hiComponentId, lowTopMostParentId, hiTopMostParentId, touchEvent) {
            callbackFunction(lowComponentId, hiComponentId, lowTopMostParentId, hiTopMostParentId, touchEvent);
        }, "vjji");
        // Init component touch callback; mouse clicks are translated into touch.
        Module.ccall("setOnComponentTouchedCallback", null, ["number"], [this.m_TouchedFuncPtr]);
    };

    /// Subscribe to the OnComponentSelectionChanged callback.
    this.SetOnComponentSelectionChanged = function (callbackFunction)
    {
        if (this.m_ComponentSelectionPtr !== null)
        {
            removeFunction(this.m_ComponentSelectionPtr);
        }

        this.m_ComponentSelectionPtr = addFunction(function (componentNamePtr)
        {
            if (componentNamePtr == 0)
            {
                return;
            }
            callbackFunction(Module.UTF16ToString(componentNamePtr));
        }, "vjji");
        // init component touch callback; mouse clicks are translated into touch
        Module.ccall("setOnComponentSelectionChangedCallback", null, ["number"], [this.m_ComponentSelectionPtr]);
    };

    /// Subscribe to be notified when the project engine state changes.
    this.SetOnProjectStateChanged = function (callbackFunction)
    {
        this.m_OnProjectStateChanged = callbackFunction;
    };

    /// Subscribe to be notified when the project is initialised.
    this.SetOnProjectInitialised = function (callbackFunction)
    {
        this.m_OnProjectInitialised = callbackFunction;
    };

    /// Subscribe to be notified when the connection has changed.
    this.SetOnConnectionChanged = function (callbackFunction)
    { 
        this.m_OnConnectionChanged   = callbackFunction;
    };

    /// Render as fast as we can manage.
    this.RenderAsFastAsPossible = true;
    /// Refresh rate for when RenderAsFastAsPossible is false.
    this.RenderingRefreshRate = 10; // frames per second

    /// If set to true, more logging information will be sent to the JavaScript console.
    this.EnableConsoleLogs = function (enable)
    {
        this.m_ConsoleLogsEnabled = enable;
    };

    this.EnableLogs = function (logId, enable)
    {
        Module.ccall("enableLogs", null, ["number", "bool"], [logId, enable]);
    };

    /// Enable logs from the EngineFront component.
    this.EnableLogsEngineFront = function (enable)
    {
        this.EnableLogs(0, enable);
    };

    /// Enable logs from the Renderer component.
    this.EnableLogsRenderer = function (enable)
    {
        this.EnableLogs(1, enable);
    };

    /// Enable logs from the ClientModuleManager component.
    this.EnableLogsClientModuleManager = function (enable)
    {
        this.EnableLogs(2, enable);
    };

    /// Resize the canvas to fill its parent. Call on document.body.onresize.
    this.ResizeCanvas = function ()
    {
        let canvas = document.getElementById("canvas");

        if (canvas !== null)
        {
            // Make it visually fill the positioned parent
            canvas.style.width = "100%";
            canvas.style.height = "100%";

            // ...then set the internal size to match
            canvas.width = canvas.offsetWidth;
            canvas.height = canvas.offsetHeight;
        }
    };

    /// Get an integer value representing the EngineState.
    this.GetState = function ()
    {
        return Module.ccall("getEngineState", "number");
    };

    /// Set the current camera by giving a camera name.
    this.SetCamera = function (cameraName)
    {
        this.LOG("Change Camera: " + cameraName);

        // Convert JavaScript string to UTF-8.
        let len = cameraName.length + 1;
        let ptr = stackAlloc(len);
        stringToUTF8(cameraName, ptr, len);
        Module.ccall('setCamera', null, ["number", "bool"], [ptr, true]);
    };

    /// Return an array of strings containing every camera's name.
    this.GetCameraNames = function ()
    {
        let str   = new Sym3String(2048);
        let len   = Module.ccall("getCameraNames", "number", ["number", "number"], [str.GetPtr(), str.GetLength()]);
        let text  = this._DecodeUtf16String(str.GetView(), len);
        str.Free();
        let names = text.split("\n");
        if (names[names.length - 1].length == 0)
        {
            return names.slice(0, names.length - 1);
        }
        return names;
    };

    /// Return an array of strings containing every component's name.
    this.GetComponentNames = function ()
    {
        let componentNames = Module.ccall("getComponentNames", "number", [], []);
        let namesString = Module.UTF16ToString(componentNames);
        Module.ccall("freeWStringBuffer", null, ["number"], [componentNames]);
        let names = namesString.split("\n");

        if (names[names.length - 1].length == 0)
        {
            return names.slice(0, names.length - 1);
        }
        return names;
    };

    /// Animate an equipment's title.
    this.AnimateEquipmentTitle = function (lowComponentId, hiComponentId)
    {
        Module.ccall("animateTitleModelSize", null, ["number", "number"], [lowComponentId, hiComponentId]);
    };

    /// Get a property value from the server.
    this.GetServerProperty = function (propertyName)
    {
        let propertyStr = new Sym3String(propertyName);
        let textBuffer  = new Sym3String(1024);

        let len  = Module.ccall("getServerProperty", "number", ["number", "number", "number"], [propertyStr.GetPtr(), textBuffer.GetPtr(), textBuffer.GetLength()]);
        let text = this._DecodeUtf16String(textBuffer.GetView(), len);

        textBuffer.Free();
        propertyStr.Free();

        return text;
    };

    /// Set the product display filter string.
    this.SetProductUserPropertyDisplay = function (filterString)
    {
        if (this.m_LastEngineState < Sym3EngineState.Active)
        {
            return;
        }

        let buffer = new Sym3String(filterString);
        Module.ccall("setProductUserPropertyDisplay", null, ["number"], [buffer.GetPtr(), true]);
        buffer.Free();
    };

    /// Change the background color. Takes r, g, b as byte values (0-255).
    this.ChangeBackgroundColor = function (red, green, blue)
    {
        Module.ccall("setBackgroundColor", null, ["number", "number", "number"], [red, green, blue]);
    };

    /// Reset to the default view.
    this.ResetView = function ()
    {
        _gotoHomePosition();
    };

    /// Toggle between 2D (orthogonal top down view) and 3D.
    this.Toggle2D3D = function ()
    {
        _togglePerspectiveProjection();
    };

    /// Stop rendering the 3D scene.
    this.DisableRendering = function ()
    {
        _disableRendering();
    };

    /// Start rendering the 3D scene.
    this.EnableRendering = function ()
    {
        _enableRendering();
    };

    /// Set whether animations are enabled (true) or disabled (false).
    this.EnableAnimations = function (enable)
    {
        Module.ccall("setRenderAnimationsEnabled", null, ["bool"], [enable]);
    };

    /// Make a component the center of the screen, without affecting the zoom level.
    this.CenterComponent = function(componentName)
    {
        let component = new Sym3String(componentName);
        let success = Module.ccall("centreComponent", "bool", ["number"], [component.GetPtr()]);
        component.Free();
        return success;
    };

    /// Zoom to a component of the given name. Returns true on success.
    this.ZoomToComponent = function (componentName)
    {
        let buffer = new Sym3String(componentName);
        let success = Module.ccall("zoomToComponent", "bool", ["number"], [buffer.GetPtr()]);
        buffer.Free();
        return success;
    };

    /**
     * Disconnect from the server.
     * 
     * If there is no connection, this does nothing.
     * Call this before making a new connection with Connect.
     */
    this.Disconnect = function ()
    {
        this.sym3Socket.Disconnect();
    }

    this.Connect = function (port, serverName)
    {
        if (serverName == null)
        {
            serverName = location.hostname;
        }

        this.sym3Socket.Connect2(port, serverName);
    }

    this.AttachTooltip = function (componentName, text)
    {
        let buffer = new Sym3String(componentName);
        let ctext  = new Sym3String(text);
        Module.ccall("attachTooltip", null, ["number", "number"], [buffer.GetPtr(), ctext.GetPtr()])
        ctext.Free();
        buffer.Free();
    };

    this.RemoveTooltip = function (componentName)
    {
        let component = new Sym3String(componentName);
        Module.ccall("removeTooltip", null, ["number"], [component.GetPtr()]);
        component.Free();
    }

    // Private interfaces follow.

    /// Get the number of messages handled.
    this.GetStats = function ()
    {
        return Module.ccall("externalDriverGetNumberOfMessageHandled", "number");
    };

    this.LOG = function (msg)
    {
        if (this.m_ConsoleLogsEnabled)
        {
            console.log(`Sym3 ==> ${msg}`);
        }
    };

    this.ERROR = function (msg)
    {
        console.error(`Sym3-ERROR ==> ${msg}`);
    };

    this._DecodeUtf16String = function (textArray, len)
    {
        if (len == 0)
        {
            return "";
        }

        let text = "";
        for (let i = 0; i < len; i++)
        {
            text = text + String.fromCharCode(textArray[i]);
        }
        return text;
    };

    this._MonitorEngineState = function ()
    {
        let state = this.GetState();
        if ((state != this.m_LastEngineState) || (state < Sym3EngineState.AwaitingConnectionParameters))
        {
            if (this.m_OnProjectStateChanged != null)
            {
                this.m_OnProjectStateChanged(state);
            }
            if (state >= Sym3EngineState.Active)
            {
                this.LOG(`Startup took ${performance.now() - timeAtStart} milliseconds.`);
                timeAtStart = performance.now();

                let cameraNames = this.GetCameraNames();
                let projectName = this.GetServerProperty("ProjectName");
                if (!projectName)
                {
                    projectName = "Not defined";
                }
                let projectGuid     = this.GetServerProperty("ProjectGuid");
                let protocolVersion = this.GetServerProperty("ProtocolVersion");
                let stateVersion    = this.GetServerProperty("StateVersion");

                if (this.m_OnProjectInitialised != null)
                {
                    this.m_OnProjectInitialised(projectName, projectGuid, protocolVersion, stateVersion, cameraNames);
                }

                this.EnableRendering();
            }
            this.m_LastEngineState = state;
        }
        if (state == Sym3EngineState.AwaitingConnection)
        {
            this.sym3Socket.Connect();
        }
    };

    /// Send data, if the engine has any to send.
    this._SendAnyData = function ()
    {
        let len = this.m_ExternalDriverGetDataToSend(this.m_SendBuffer, this.m_SendBufferSize);
        if (len > 0)
        {
            // We can't cache this TypedArray -- if Emscripten grows memory, it will invalidate it.
            let data = new Uint8Array(Module.HEAPU8.buffer, this.m_SendBuffer, this.m_SendBufferSize);
            let dataToSend = data.slice(0, len);
            this.sym3Socket.Send(dataToSend);
        }
        else if (len < 0)
        {
            console.error('sendBuffer too small');
        }
    };

    /// Get a component id from a component name.
    this._GetComponentIdByName = function (componentName)
    {
        let buffer = new Sym3String(componentName);
        let idname = Module.ccall("getComponentIdByName", "number", ["number"], buffer.GetPtr());
        let id     = Module.Pointer_stringify(idname)
        Module.ccall("freeStringBuffer", null, ["number"], [idname]);
        buffer.Free();
        return id;
    };

    /// Get a component name from a component ID.
    this._GetComponentNameById = function (componentId)
    {
        let compNames = Module.ccall("getComponentNameById", "number", ["string"], [componentId + ""]);
        let str = Module.UTF16ToString(compNames);
        Module.ccall('freeWStringBuffer', null, ["number"], [compNames]);
        return str;
    };

    this._Update = function ()
    {
        _runEngineOnce();
        this._SendAnyData();
        Module.ccall("runAndRenderFrame", null, ["bool"], [false]);
        _justRenderFrame();

        if (this.RenderAsFastAsPossible)
        {
            window.requestAnimationFrame(this._Update.bind(this));
        } else
        {
            // Using interval.
            window.setTimeout(redraw, 1000 / this.RenderingRefreshRate);
        }
    };

    this._Start = function ()
    {
        _runSym3OperatorRuntime();
        this.LOG("Sym3 started");
        this.m_ExternalDriverGetDataToSend = Module.cwrap('externalDriverGetDataToSend', "number", ["number", "number"]);
        setInterval(function () {SYM3._MonitorEngineState(); }, 2000);
    };

    this._Run = function ()
    {
        this.ResizeCanvas();

        // Allocate a small buffer for the WebSocket.
        this.m_SendBuffer = Module._malloc(this.m_SendBufferSize);

        this.sym3Socket.Create();

        this._Init(this.m_ClientCategories);

        this.EnableLogsEngineFront(false);
        this.EnableLogsRenderer(false);
        this.EnableLogsClientModuleManager(false);

        this._CreateView();

        this._Start();

        this.DisableRendering(); // Reenabled in MonitorEngine state once everything is up and running.
        this._Update();
    };

    this._Init = function (clientCategories)
    {
        // There are no parameter options and null/zero is passed in most cases. The compiled code sets parameters appropriately.
        Module.ccall("initSym3OperatorRuntime", null, ["number", "number", "number", "number"], [0, clientCategories, 0, 0]);
        Module.ccall("setIsUsingExternalDriver", null, ["bool"], [true]);
        this.LOG("Sym3 initialized.");
    };

    this._CreateView = function ()
    {
        Module.ccall("addView", null, ["number"], [0]);
        this.LOG("3D view created.");
    };

    this._NbBytesReceived = 0;
    this._NbMessagesReceived = 0;

    this._UpdateBytesReceived = function (nbBytes)
    {
        this._NbBytesReceived = this._NbBytesReceived + nbBytes;
        this._NbMessagesReceived++;
    };
}

var SYM3;
function CreateSym3(webSocketPort, serverName)
{
    let clientCategories = 8 | 16 | 32 | 64 | 128 | 256;

    if (webSocketPort == null)
    {
        webSocketPort = 27000; // try with default one
    }

    // Module is globally defined in the sym3operatorruntime.js code.
    Module ={
        noInitialRun: true,
        print: (function () {
            return function (text)
            {
                if (arguments.length > 1)
                {
                    text = Array.prototype.slice.call(arguments).join(' ');
                }
                SYM3.LOG(text);
            };
        })(),
        printErr: function (text)
        {
            if (arguments.length > 1)
            {
                text = Array.prototype.slice.call(arguments).join(' ');
            }
            else
            {
                console.error(text);
            }
        },
        locateFile: function (filename)
        {
            return `sym3/${filename}`;
        },
        canvas: (function ()
        {
            let canvas = document.getElementById("canvas");

            if (canvas !== null)
            {
                canvas.addEventListener("webglcontextlost", function (e)
                {
                    alert("WebGL context lost. You will need to reload the page.");
                    e.preventDefault();
                }, false);
            }

            return canvas;
        })(),
        setStatus: function (text)
        {
            if (!Module.setStatus.last)
            {
                Module.setStatus.last = {
                    time: Date.now(),
                    text: ""
                };
            }
            if (text === Module.setStatus.text)
            {
                return;
            }
            var m = text.match(/([^(]+)\((\d+(\.\d+)?)\/(\d+)\)/);
            var now = Date.now();
            if (m && now - Date.now() < 30)
            {
                return; // if this is a progress update, skip it if too soon
            }
            if (m)
            {
                text = m[1];
            }
            else
            {
                if (!text)
                {
                    SYM3._Run();
                }
            }
        },
        totalDependencies: 0,
        monitorRunDependencies: function (left)
        {
            this.totalDependencies = Math.max(this.totalDependencies, left);
            Module.setStatus(left ? `Preparing... (${this.totalDependencies - left}/${this.totalDependencies})` : "All downloads complete.");
        }
    };

    SYM3 = new Sym3RuntimeAPI(clientCategories, serverName, webSocketPort);
    self.SYM3 = SYM3;
    Module.setStatus("Downloading...");
    window.onerror = function (event)
    {
        // TODO: do not warn on ok events like simulating an infinite loop or exitStatus
        Module.setStatus("Exception thrown, see JavaScript console.");
        Module.setStatus = function (text)
        {
            if (text)
            {
                Module.printErr(`[post-exception status] ${text}`);
            }
        };
    };

    return SYM3;
}

function redraw()
{
    SYM3._Update();
}

var g_FileReader = null;
/// Wrapper around the WebSocket that communicates with the Sym3 server.
function Sym3WebSocket(serverName, webSocketPort)
{
    this.m_ConsoleLogsEnabled = true;
    this.m_WebSocket = null;
    this.m_ServerName = serverName;
    this.m_WebSocketPort = webSocketPort;
    this.m_LastAwaitingEngineTime = new Date();

    this.Create = function ()
    {
        this.m_WebSocket = this.CreateWebSocket();
    };

    this.Connect = function ()
    {
        if ((!this.m_WebSocket) || (this.m_WebSocket.readyState === this.m_WebSocket.CLOSED) || (this.m_WebSocket.readyState === this.m_WebSocket.CLOSING))
        {
            this.LOG("need new websocket");
            this.Create();
        }
        else if (new Date() - this.m_LastAwaitingEngineTime > 10000)
        {
            // timeout must be shorter than backend change from state 1 back to state 0
            this.LOG("timeout - reconnecting...");
            this.Disconnect();
            this.Create();
        }
    };

    this.Connect2 = function (port, serverName)
    {
        this.m_ServerName = serverName;
        this.m_WebSocketPort = port;
        this.Connect();
    };

    this.Disconnect = function ()
    {
        if (this.m_WebSocket)
        {
            this.m_WebSocket.close();
            this.m_WebSocket = null;
        }
    };

    this.Send = function (dataToSend)
    {
        this.m_WebSocket.send(dataToSend);
    };

    this.LOG = function (msg)
    {
        if (this.m_ConsoleLogsEnabled == true)
        {
            console.log(`Sym3WebSocket ==> ${msg}`);
        }
    };

    this.ERROR = function (msg)
    {
        console.error(`Sym3WebSocket-ERROR ==> ${msg}`);
    };

    this.CreateWebSocket = function ()
    {
        let host = `ws://${this.m_ServerName}:${this.m_WebSocketPort}`;
        this.LOG(`Create WebSocket to ${host}`);
        let ws = new WebSocket(host);
        ws.binaryType = "arraybuffer";

        if ("WebSocket" in window)
        {
            this.LOG("WebSocket is supported");

            ws.onopen = function ()
            {
                _externalDriverConnect();
 
                if (SYM3.m_OnConnectionChanged   != null)
                { 
                    SYM3.m_OnConnectionChanged  ("open");
                }
            };

            ws.onmessage = function (e)
            {
                let recvarr = new Uint8Array(e.data);
                let recvbuf = Module._malloc(e.data.byteLength);
                Module.HEAPU8.set(recvarr, recvbuf);
                Module.ccall("externalDriverReceivePacket", null, ["number", "number"], [recvbuf, e.data.byteLength]);
                SYM3._UpdateBytesReceived(e.data.byteLength);
                Module._free(recvbuf);
            };

            ws.onclose = function ()
            {
                SYM3.LOG("Connection is closed...");
                _externalDriverDisconnect();
                SYM3.LOG("externalDriverDisconnect called...");
 
                if (SYM3.m_OnConnectionChanged   != null)
                { 
                    SYM3.m_OnConnectionChanged  ("closed");
                }
            };

            window.onbeforeunload = function (event)
            {
                _externalDriverDisconnect();
                ws.close();
            };
        }
        else
        {
            this.ERROR("WebSocket NOT supported by your Browser!");
        }

        return ws;
    };
}
